import lnschema_core as core
from lamin_logger import logger
from lndb import settings as setup_settings

from ._settings import settings
from .dev._core import filepath_from_dobject
from .dev.file import load_to_memory


def populate_runin(dobject: core.DObject, run: core.Run):
    setup_settings.instance._cloud_sqlite_locker.lock()
    with setup_settings.instance.session() as ss:
        result = ss.get(core.link.RunIn, (run.id, dobject.id))
        if result is None:
            ss.add(
                core.link.RunIn(
                    run_id=run.id,
                    dobject_id=dobject.id,
                )
            )
            ss.commit()
            logger.info(f"Added dobject ({dobject.id}) as input for run ({run.id}).")
            setup_settings.instance._update_cloud_sqlite_file()
    setup_settings.instance._cloud_sqlite_locker.unlock()


# this is exposed to the user as DObject.load
def load(dobject: core.DObject, stream: bool = False, is_run_input: bool = False):
    if stream and dobject.suffix not in (".h5ad", ".zarr"):
        logger.warning(f"Ignoring stream option for a {dobject.suffix} object.")

    filepath = filepath_from_dobject(dobject)
    # TODO: better design to track run inputs
    if settings.track_run_inputs_upon_load or is_run_input:
        from lamindb import nb

        if nb.run is None:
            logger.warning(
                "Input tracking for runs through `load` is currently only implemented"
                " for notebooks."
            )
        else:
            populate_runin(dobject, nb.run)
    # TODO: enable track usage
    # track_usage(dobject.id, "load")
    return load_to_memory(filepath, stream=stream)



from functools import partial
from pathlib import Path
from typing import List, Union, overload  # noqa

import sqlmodel as sqm
from lamin_logger import logger
from lndb import settings as setup_settings
from lndb.dev import UPath
from lnschema_core import DObject
from sqlalchemy.orm.attributes import set_attribute

from .._docs import doc_args
from ..file import store_file, write_adata_zarr
from ..file._file import print_hook
from ._core import dobject_to_sqm, get_session_from_kwargs
from ._select import select

add_docs = """
Insert or update data records.

Inserts a new :term:`record` if the corresponding row doesn't exist.
Updates the corresponding row with the record if it exists.

To update a row, query it with `.get` or `.select` and modify it before
passing it to `add`.

Guide: :doc:`/guide/add-delete`.

Example:

>>> # add a record (by passing a record)
>>> ln.add(wetlab.Experiment(name="My test", biometa_id=test_id))
>>> # update an existing record
>>> experiment = ln.select(wetlab.Experiment, id=experiment_id).one()
>>> experiment.name = "New name"
>>> ln.add(experiment)
>>> # add a record by fields if not yet exists
>>> ln.add(wetlab.Experiment, name="My test", biometa_id=test_id)

Args:
    record: One or multiple records as instances of `SQLModel`.
"""


@overload
def add(record: sqm.SQLModel) -> sqm.SQLModel:
    ...


# Currently seeing the following error without type ignore:
# Overloaded function signature 2 will never be matched: signature 1's parameter
# type(s) are the same or broader
@overload
def add(records: List[sqm.SQLModel]) -> List[sqm.SQLModel]:  # type: ignore
    ...


@overload
def add(  # type: ignore
    entity: sqm.SQLModel, **fields
) -> Union[sqm.SQLModel, List[sqm.SQLModel]]:
    ...


@doc_args(add_docs)
def add(  # type: ignore
    record: Union[sqm.SQLModel, List[sqm.SQLModel]], **fields
) -> Union[sqm.SQLModel, List[sqm.SQLModel]]:
    """{}"""  # noqa
    session = get_session_from_kwargs(fields)
    if isinstance(record, list):
        records = record
    elif isinstance(record, sqm.SQLModel):
        records = [record]
    else:
        model = dobject_to_sqm(record)
        results = select(model, **fields).one_or_none()
        if results is None:
            records = [model(**fields)]
        else:
            logger.info("An existing record is found in the DB:")
            return results

    if session is None:  # assume global session
        session = setup_settings.instance.session()
        setup_settings.instance._cloud_sqlite_locker.lock()
        close = True
    else:
        close = False

    # commit metadata to database
    db_error = None
    for record in records:
        prepare_filekey_metadata(record)
        session.add(record)
    try:
        session.commit()
    except Exception as e:
        db_error = e

    # upload data objects to storage
    added_records = []
    if db_error is None:
        added_records, upload_error = upload_committed_records(records, session)

    if close:
        session.close()
        setup_settings.instance._update_cloud_sqlite_file()
        setup_settings.instance._cloud_sqlite_locker.unlock()

    error = db_error or upload_error
    if error is not None:
        error_message = prepare_error_message(records, added_records, error)
        raise RuntimeError(error_message)
    elif len(added_records) > 1:
        return added_records
    else:
        return added_records[0]


def upload_committed_records(records, session):
    """Upload records in a list of database-commited records to storage.

    If any upload fails, subsequent records are cleaned up from the DB.
    """
    # make sure ALL records are up-to-date to enable accurate comparison
    # during metadata cleanup
    error = None
    added_records = []
    for record in records:
        session.refresh(record)

    # upload data objects
    for record in records:
        if isinstance(record, DObject) and hasattr(record, "_local_filepath"):
            try:
                upload_data_object(record)
            except Exception as e:
                error = e
                break
        added_records += [record]

    # clean up metadata for objects not uploaded to storage
    if error is not None:
        for record in records:
            if record not in added_records:
                session.delete(record)
        session.commit()

    return added_records, error


def prepare_error_message(records, added_records, error) -> str:
    if len(records) == 1 or len(added_records) == 0:
        error_message = (
            "An error occured. No entries were uploaded or committed"
            " to the database. See error message below.\n\n"
        )
    else:
        error_message = (
            "An error occured. The following entries have been"
            " successfully uploaded and committed to the database:\n"
        )
        for record in added_records:
            error_message += (
                f"- {', '.join(record.__repr__().split(', ')[:3]) + ', ...)'}\n"
            )
        error_message += "\nSee error message below.\n\n"
    error_message += f"{str(error)}"
    return error_message


def prepare_filekey_metadata(record) -> None:
    """For cloudpath, write custom filekey to _filekey.

    A filekey excludes the storage root and the file suffix.
    """
    storage = setup_settings.instance.storage

    def set_filekey(record: sqm.SQLModel, filepath: Union[Path, UPath]):
        filepath_str = filepath.as_posix()
        root_str = storage.root.as_posix()
        if root_str[-1] != "/":
            root_str += "/"
        set_attribute(
            record,
            "_filekey",
            filepath_str.replace(root_str, "").replace(record.suffix, ""),
        )

    # _local_filepath private attribute is only added
    # when creating DObject from data
    if isinstance(record, DObject) and hasattr(record, "_local_filepath"):
        if record.suffix == ".zarr":
            pass
        else:
            # cloud storage
            if record._cloud_filepath is not None:
                set_filekey(record, record._cloud_filepath)
            # local storage that is configured
            else:
                if storage.root in record._local_filepath.parents:
                    set_filekey(record, record._local_filepath)


def upload_data_object(dobject) -> None:
    """Store and add dobject and its linked entries."""
    dobject_storage_key = f"{dobject.id}{dobject.suffix}"

    storage = setup_settings.instance.storage

    if dobject.suffix != ".zarr":
        # no file upload for cloud storage or configured local storage
        # only copy data when file is not in the configured local storage
        if (dobject._cloud_filepath is None) and (
            storage.root not in dobject._local_filepath.parents
        ):
            store_file(dobject._local_filepath, dobject_storage_key)
    else:
        storagepath = storage.key_to_filepath(dobject_storage_key)
        print_progress = partial(print_hook, filepath=dobject.name)
        write_adata_zarr(dobject._memory_rep, storagepath, callback=print_progress)



from pathlib import Path
from typing import List, Optional, Union

import nbproject as _nb
from lamin_logger import logger
from lndb import settings
from lnschema_core import Notebook, Run, dev


class nb:
    """Manage Jupyter notebooks.

    For more background, see `nbproject <https://lamin.ai/docs/nbproject>`__.
    """

    notebook: Notebook = None  # Notebook of this Python session
    run: Run = None  # run of this Python session

    @classmethod
    def header(
        cls,
        *,
        run: Optional[str] = None,
        pypackage: Union[str, List[str], None] = None,
        filepath: Optional[str] = None,
        env: Optional[str] = None,
        id: Optional[str] = None,
        v: Optional[str] = "0",
        name: Optional[str] = None,
    ):
        """Track the notebook & display metadata.

        Call without arguments in most settings.

        If the notebook has no nbproject metadata, initializes & writes metadata
        to disk.

        Args:
            run: If `None`, loads the latest run of the notebook (or creates on if
                there isn't any).

                If `"new"`, instantiates a new `Run` record.

            pypackage: One or more python packages to track.
            filepath: Filepath of notebook. Only needed if automatic inference fails.
            env: Editor environment. Only needed if automatic inference fails.
                Pass `'lab'` for jupyter lab and `'notebook'` for jupyter notebook,
                this can help to identify the correct mechanism for interactivity
                when automatic inference fails.
            id: Pass a notebook id manually.
            v: Pass a notebook version manually.
            name: Pass a notebook name manually.
        """
        if id is None and name is None:
            nbproject_failed_msg = (
                "Auto-retrieval of notebook name & title failed.\nPlease paste error"
                " at: https://github.com/laminlabs/nbproject/issues/new \n\nFix: Run"
                f" ln.nb.header(id={dev.id.notebook()}, name='my-notebook-name')"
            )
            try:
                _nb.header(pypackage=pypackage, filepath=filepath, env=env)
            except Exception:
                raise RuntimeError(nbproject_failed_msg)
            # this contains filepath if the header was run successfully
            from nbproject._header import _filepath

            id = _nb.meta.store.id
            v = _nb.meta.store.version
            name = Path(_filepath).stem
            title = _nb.meta.live.title
        elif id is None or name is None:
            # Both id and name need to be passed if passing it manually
            raise RuntimeError("Fix: Pass both id & name to ln.nb.header().")
        else:
            title = None

        logger.info(f"Instance: {settings.instance.owner}/{settings.instance.name}")

        import lamindb as ln
        import lamindb.schema as lns

        notebook = ln.select(
            lns.Notebook,
            id=id,
            v=v,
        ).one_or_none()
        if notebook is None:
            notebook = lns.Notebook(
                id=id,
                v=v,
                name=name,
                title=title,
            )
            notebook = ln.add(notebook)
            logger.info(f"Added notebook: {notebook.id} v{notebook.v}")
        else:
            logger.info(f"Loaded notebook: {notebook.id} v{notebook.v}")
            if notebook.name != name or notebook.title != title:
                notebook.name = name
                notebook.title = title
                ln.add(notebook)
                logger.info("Updated notebook name or title.")

        # at this point, we have a notebook object
        cls.notebook = notebook

        # check user input
        # if isinstance(run, lns.Run):
        # This here might be something we may want in the future
        # but checking all the cases in which that run record has integrity
        # is quite a bit of code - not now!
        #     run_test = ln.select(lns.Run, id=run.id).one_or_none()
        #     if run_test is None:
        #         logger.info("Passed run does not exist, adding it")
        #         ln.add(run)
        if run is None:
            # retrieve the latest run
            run = (
                ln.select(lns.Run, notebook_id=notebook.id, notebook_v=notebook.v)
                .order_by(lns.Run.created_at.desc())
                .first()
            )
            if run is not None:
                logger.info(f"Loaded run: {run.id}")  # type: ignore
        elif run != "new":
            raise ValueError("Fix: ln.nb.header(run='new')!")

        # create a new run if doesn't exist yet or is requested by the user ("new")
        if run is None or run == "new":
            run = lns.Run(notebook_id=notebook.id, notebook_v=notebook.v)
            run = ln.add(run)  # type: ignore
            logger.info(f"Added run: {run.id}")  # type: ignore

        # at this point, we have a run object
        cls.run = run

    @classmethod
    def publish(cls, version: str = None, i_confirm_i_saved: bool = False):
        """Publish the notebook.

        Args:
            version: Notebook version to publish.
            i_confirm_i_saved: Only relevant outside Jupyter Lab & Notebook as a
                safeguard against losing the editor buffer content.
        """
        from nbproject._publish import finalize_publish, run_checks_for_publish

        import lamindb as ln
        import lamindb.schema as lns

        result = run_checks_for_publish(
            calling_statement="publish(", i_confirm_i_saved=i_confirm_i_saved
        )
        if result != "checks-passed":
            return result
        finalize_publish(calling_statement="publish(", version=version)
        # update DB
        notebook = ln.select(Notebook, id=cls.notebook.id, v=cls.notebook.v).one()
        # update version
        notebook.title = _nb.meta.live.title
        if version != cls.notebook.v:
            notebook_add = lns.Notebook(
                id=notebook.id, v=version, name=notebook.name, title=notebook.title
            )
        else:
            notebook_add = notebook
        ln.add(notebook_add)
        if version != cls.notebook.v:
            cls.run.notebook_v = version
            ln.add(cls.run)
            ln.delete(notebook)



from bionty import __version__ as bionty_v
from lndb import __version__ as lndb_v
from lnschema_core import __version__ as lnschema_core_v
from nbproject import __version__ as nbproject_v
from packaging import version

if version.parse(lndb_v) != version.parse("0.35.rc1"):
    raise RuntimeError("Upgrade lndb! pip install lndb==0.35.rc1")

if version.parse(lnschema_core_v) != version.parse("0.27.2"):
    raise RuntimeError("lamindb needs lnschema_core==0.27.2")

if version.parse(bionty_v) != version.parse("0.7.0"):
    raise RuntimeError("lamindb needs bionty==0.7.0")

if version.parse(nbproject_v) < version.parse("0.8.2"):
    raise RuntimeError("lamindb needs nbproject>=0.8.2")



import nox
from lndb.test.nox import build_docs, login_testuser1, run_pre_commit, run_pytest

nox.options.reuse_existing_virtualenvs = True


@nox.session(python=["3.7", "3.8", "3.9", "3.10", "3.11"])
def lint(session: nox.Session) -> None:
    run_pre_commit(session)


@nox.session(python=["3.7", "3.8", "3.9", "3.10", "3.11"])
def build(session):
    login_testuser1(session)
    login_user_2 = "lndb login testuser2@lamin.ai --password goeoNJKE61ygbz1vhaCVynGERaRrlviPBVQsjkhz"  # noqa
    session.run(*(login_user_2.split(" ")), external=True)
    session.install(".[dev,test]")
    run_pytest(session)
    build_docs(session)



"""LaminDB: Manage data & analyses.

Import the package::

   import lamindb as ln
   import lamindb.schema as lns

The central data object, a wrapper for files, on-disk (`zarr`, etc.) and
in-memory objects (`DataFrame`, `AnnData`, etc.):

.. autosummary::
   :toctree: .

   DObject

Query & manipulate data:

.. autosummary::
   :toctree: .

   select
   add
   delete

Manipulate data with open session:

.. autosummary::
   :toctree: .

   Session

View DB content:

.. autosummary::
   :toctree: .

   view

Schema - entities and their relations:

.. autosummary::
   :toctree: .

   schema

Manage knowledge:

.. autosummary::
   :toctree: .

   knowledge

Track Jupyter notebooks:

.. autosummary::
   :toctree: .

   nb

Developer API:

.. autosummary::
   :toctree: .

   settings
   dev
"""

__version__ = "0.28rc1"  # denote a release candidate for 0.1.0 with 0.1rc1

# prints warning of python versions
from lamin_logger import py_version_warning

py_version_warning("3.8", "3.10")

from lndb import settings as _setup_settings
from lndb._migrate import check_deploy_migration as _check_migrate

from . import _check_versions  # executes checks during import

if _setup_settings.instance.storage.root is None:
    raise RuntimeError(
        "Please run `lndb load` or `lndb init` to configure an instance."
    )
_check_migrate(usettings=_setup_settings.user, isettings=_setup_settings.instance)

from lnschema_core import DObject  # noqa

from . import dev  # noqa
from . import knowledge  # noqa
from . import schema  # noqa
from ._delete import delete  # noqa
from ._nb import nb  # noqa
from ._settings import settings
from ._subset import subset
from ._view import view  # noqa
from .dev.db import Session  # noqa
from .dev.db._add import add  # noqa
from .dev.db._select import select  # noqa
from .dev.object._lazy_field import lazy
